/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
******************************************************************************
**                                                                            
**        Y     Y   OOOO    OOOO   DDDDD     AA
**         Y   Y   O    O  O    O  D    D   A  A
**          Y Y    O    O  O    O  D    D  A    A 
**           Y     O    O  O    O  D    D  AAAAAA
**           Y     O    O  O    O  D    D  A    A
**           Y      OOOO    OOOO   DDDDD   A    A
**
******************************************************************************
******************************************************************************/
/*****************************************************************************
*
*  MODULE :             Collections
*  SCOPE :              Collections common services
*  FILE :               col.cc
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/

#pragma implementation "colelem.hxx"
#pragma implementation "colroot.hxx"
#pragma implementation "colln.hxx"
#pragma implementation "col.hxx"

/* IMPORTS *******************************************************************/

#include <string.h>

#ifndef __col_HXX
#include "col.hxx"
#endif

#ifndef __colelem_HXX
#include "colelem.hxx"
#endif

#ifndef __yooexcept_HXX
#include "yooexcept.hxx"
#endif

/*****************************************************************************/

/*= EXTERNAL FUNCTION =======================================================*/
const yoo_ObjectX& Collection::max()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return NULL_REF ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int Collection::indKey(const void* key,
		       void* (*)(const void*),int (*)(const void*,const void*))
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return -1 ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<Collection> Collection::insertDirect(const void* atom, int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int currElem = this->nbElem() ;
  int prev = (pos == COL_TAIL) ? currElem-1 : pos-1 ;
  // Leaf-node is full => split in two parts
  if (this->maxElem() == currElem) {
    return this->split(atom, prev) ;
  } else {
    // insertion at the end need no move
    if (pos != COL_TAIL) {
      memmove(this->tab(pos+1),this->tab(pos),this->sizeElem()*(currElem-pos));
    }
    this->newChild(atom, prev+1) ;
    return NULL ;
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
void Collection::removeDirect(int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->delChild(pos) ;
  int currElem = this->nbElem() ;
  // remove at the end need no move
  if (pos < currElem) {
    memmove(this->tab(pos),this->tab(pos+1),this->sizeElem()*(currElem-pos));
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<Collection> Collection::split(const void* atom, int prev)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int flag = FALSE ;
  Collection* ln = this->clone() ;
  int max = this->nbElem() ;

  // move last half of the leaf-node into the clone
  for (int curr = max / 2 ; curr < max; curr++) {
    ln->insertDirect(this->tab(curr)->data(), COL_TAIL) ;
    // test if the new element must be inserted in this leaf-node
    if (curr == prev) {
      ln->insertDirect(atom, COL_TAIL) ;
      flag = TRUE ;
    }
    this->delChild(curr) ;
  }
  
  // test if the new element has been inserted in new leaf-node  
  if (flag == FALSE) {
    this->insertDirect(atom, prev+1) ;
  }
  return ln ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<ColLeaf> ColLeaf::getLeaf(int)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return this ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void ColLeaf::newChild(const void* atom, int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  memcpy(this->tab(pos)->data(), atom, sizeof(ColElem)) ;
  this->_card++ ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void ColLeaf::delChild(int)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->_card-- ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<ColLeaf> ColLeaf::grow(int nbe)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int max = this->maxElem() ;
  // Compute number of new elements which will be created in the leaf
  nbe = ( (nbe+max) > ColNbInLeaf) ? (ColNbInLeaf - max) : nbe ;

  int ss = yoo_slotSize(sizeof(Collection)+(nbe+max)*sizeof(ColElem)) ;
  return (ColLeaf*)this->changeSize(ss) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int ColLeaf::sizeElem()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return sizeof(ColElem) ;    
}


/*= EXTERNAL FUNCTION =======================================================*/
int ColLeaf::maxElem()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return (getSlotSize()*SLOT_SIZE - sizeof(Collection)) / sizeof(ColElem) ;    
}


/*= EXTERNAL FUNCTION =======================================================*/
int ColLeaf::nbElem()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return _card ;
}


/*= EXTERNAL FUNCTION =======================================================*/
ColElem* ColLeaf::tab(int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return (ColElem*) &(this->_tab[pos*sizeof(ColElem)]) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<Collection> ColLeaf::insertTree(const void* atom, int ind)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return this->insertDirect(atom, ind) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void ColLeaf::removeTree(int ind)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->removeDirect(ind) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void ColLeaf::clear()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  delete this ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void* ColLeaf::getElem(int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return this->tab(pos)->data() ;  
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<ColLeaf> ColLeaf::scanLeaf(int ind, int& beg, int& end)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if ( (ind < 0) || (ind >= this->_card) ) {
    return NULL ;
  }
  beg = ind ;
  end = this->_card - 1 ;
  return this ;  
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<Collection> ColLeaf::insert(const void* atom, int ind)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  Ref<ColLeaf> l1 ;
  Collection* res ;

  if (this->maxElem() == this->count()) {
    l1 = this->grow(YOO_GROWING_STEP) ;
  } else {
    l1 = this ;
  }
  
  Ref<Collection> l2 = l1->insertDirect(atom, ind) ;
  if (l2 != NULL) {
    res = l1->newNode() ;
    res->insertDirect(&l1, 0) ;
    res->insertDirect(&l2, 1) ;
  } else {
    res = l1 ;
  }

  return res ;  
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<Collection> ColLeaf::remove(int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->removeDirect(pos) ;
  return this ;
}


/*= EXTERNAL FUNCTION =======================================================*/
int ColNode::node(int& ind)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int pos = 0 ;
  while (ind >= ((ColNodeElem*)this->tab(pos))->card) {
    ind -= ((ColNodeElem*)this->tab(pos++))->card ;
    if (pos >= this->nbElem()) call_throw(YOO_ERR_COL) ;    
  }
  return pos ;  
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<ColLeaf> ColNode::getLeaf(int ind)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int pos = this->node(ind) ;
  return this->tab(pos)->ref()->getLeaf(ind) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void ColNode::newChild(const void* atom, int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->tab(pos)->ref() = *((Ref<Collection>*) atom) ;
  ((ColNodeElem*)this->tab(pos))->card = this->tab(pos)->ref()->_card ;
  this->_card += ((ColNodeElem*)this->tab(pos))->card ;
  this->_nbElem += 1 ;  
}


/*= EXTERNAL FUNCTION =======================================================*/
void ColNode::delChild(int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->_card -= ((ColNodeElem*)this->tab(pos))->card ;
  this->_nbElem-- ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void ColNode::resetChild(int pos)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int diff = tab(pos)->ref()->_card - ((ColNodeElem*)this->tab(pos))->card ;
  ((ColNodeElem*)this->tab(pos))->card += diff ;
  this->_card += diff ;  
}


/*= EXTERNAL FUNCTION =======================================================*/
void ColNode::merge(int src)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int dst = (src == 0) ? src++ : src - 1 ;;
  Ref<Collection> res ;
  Ref<Collection> node = this->tab(src)->ref() ;

  this->removeDirect(src) ;

  int max = node->nbElem() ;
  for (int ptr = 0; ptr < max; ptr++) {
    res = tab(dst)->ref()->insertDirect(node->tab(ptr)->data(),COL_TAIL);
    if (res != NULL) {
      this->resetChild(dst++) ; 
      this->insertDirect(&res, dst) ;
    }
  }
  this->resetChild(dst) ; 
}


/*= EXTERNAL FUNCTION =======================================================*/
int ColNode::nbElem()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return _nbElem ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<Collection> ColNode::insertTree(const void* atom, int ind)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int ptr ;
  if (ind == COL_TAIL) {
    ind = this->_card ;
  } 
  ind--; ptr = this->node(ind) ; ind++ ;
  
  Ref<Collection> node = this->tab(ptr)->ref()->insertTree(atom, ind) ;
  this->resetChild(ptr) ;
  return (node!=NULL) ? this->insertDirect(&node,ptr+1):Ref<Collection>(NULL);
}


/*= EXTERNAL FUNCTION =======================================================*/
void ColNode::removeTree(int ind)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int ptr = this->node(ind) ;

  this->tab(ptr)->ref()->removeTree(ind) ;
  this->resetChild(ptr) ;
  if (this->tab(ptr)->ref()->nbElem() < this->tab(ptr)->ref()->maxElem()/2) {
    this->merge(ptr) ;
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
void ColNode::clear()
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  for (int i = 0; i < this->_nbElem; i++) {
    this->tab(i)->ref()->clear() ;
  }
  delete this ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void* ColNode::getElem(int ind)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  int pos = this->node(ind) ;
  return this->tab(pos)->ref()->getElem(ind) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<ColLeaf> ColNode::scanLeaf(int ind, int& beg, int& end)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  if ( (ind < 0) || (ind >= this->count()) ) {
    return NULL ;
  }

  int pos = this->node(ind) ;
  return this->tab(pos)->ref()->scanLeaf(ind, beg, end) ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<Collection> ColNode::insert(const void* atom, int ind)
/*=
  los[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  Collection* root = NULL ;
  Ref<Collection> l1 = this ;
  Ref<Collection> l2 = this->insertTree(atom, ind) ;
  if (l2 != NULL) {
    root = this->clone() ;
    root->insertDirect(&l1, 0) ;
    root->insertDirect(&l2, 1) ;
  } else {
    root = this ;
  }
  return root ;
}


/*= EXTERNAL FUNCTION =======================================================*/
Ref<Collection> ColNode::remove(int ind)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  this->removeTree(ind) ;
  if (this->_nbElem == 1) {
    Collection* l = this->tab(0)->ref() ;
    delete this ;
    return l ;
  } else {
    return this ;
  }  
}
